use std::sync::Arc;

use chrono::{DateTime, NaiveDateTime, Utc};
use config::get_config;
use error::AppResult;
use futures_util::{future::try_join_all, StreamExt};
use markets::{poloniex::PoloniexClient, Market};
use models::{Candle, CandleExtended, CandleInterval, Trade, TradeDirection};
use repos::{sqlite::SqliteRepo, Repo};

mod config;
mod error;
mod markets;
mod models;
mod repos;

#[macro_use]
extern crate async_trait;

async fn fetch_candles_until_now(
    market_client: Arc<impl Market>,
    pair: String,
    interval: CandleInterval,
    start_time: NaiveDateTime,
) -> AppResult<(Vec<Candle>, String)> {
    let mut result = vec![];
    let mut end_time = Utc::now().naive_utc();
    let mut is_completed = false;

    // max limit is 500
    let limit = 500;

    while !is_completed {
        println!(
            "{}:{}: pulling candles from [{start_time}, {end_time}]",
            pair,
            interval.as_ref()
        );
        let candles = market_client
            .get_historical_candles(&pair, interval, start_time, end_time, limit)
            .await?;

        // если какие-то свечки недополучили из-за лимитов,
        // смещаем запрашиваемый временной отрезок влево
        let time_offset = interval.as_timedelta() * limit as i32;
        end_time -= time_offset;

        if end_time < start_time {
            is_completed = true;
        }

        result.extend(candles);
    }

    println!(
        "{}:{}: pulled {} candles",
        pair,
        interval.as_ref(),
        result.len(),
    );
    Ok((result, pair.to_string()))
}

async fn calculate_new_candles(
    repo: Arc<impl Repo>,
    interval: CandleInterval,
    trade: Trade,
) -> AppResult<()> {
    let is_buy = matches!(trade.taker_side, TradeDirection::Buy);
    let insert_new_candle = || async {
        let interval_secs = interval.as_timedelta().num_seconds();
        let new_candle_ts = DateTime::from_timestamp(
            (trade.ts.and_utc().timestamp() / interval_secs) * interval_secs,
            0,
        )
        .unwrap()
        .naive_utc();

        let new_candle = CandleExtended {
            candle: Candle {
                low: trade.price,
                high: trade.price,
                open: trade.price,
                close: trade.price,
                amount: trade.amount,
                quantity: trade.quantity,
                buy_taker_amount: if is_buy { trade.amount } else { 0.0 },
                buy_taker_quantity: if is_buy { trade.quantity } else { 0.0 },
                trade_count: 1,
                ts: trade.ts,
                weighted_average: trade.amount / trade.quantity,
                interval,
                start_time: new_candle_ts,
                close_time: NaiveDateTime::UNIX_EPOCH,
            },
            pair: trade.symbol.clone(),
        };

        repo.upsert_candles(&[new_candle]).await?;

        AppResult::Ok(())
    };

    let last_candle = repo
        .get_latest_candle_from_interval(&trade.symbol, interval)
        .await?;

    if let Some(mut last_candle) = last_candle {
        let interval_delta = last_candle.candle.interval.as_timedelta();

        // если трейд не входит в интервал последней свечи, то создаём новую свечу, иначе обновляем предыдущую
        if trade.ts > (last_candle.candle.ts + interval_delta) {
            insert_new_candle().await?;
        } else {
            last_candle.candle.low = last_candle.candle.low.min(trade.price);
            last_candle.candle.high = last_candle.candle.high.max(trade.price);
            last_candle.candle.close = trade.price;
            last_candle.candle.amount += trade.amount;
            last_candle.candle.quantity += trade.quantity;
            last_candle.candle.trade_count += 1;
            last_candle.candle.ts = trade.ts;
            last_candle.candle.weighted_average =
                last_candle.candle.amount / last_candle.candle.quantity;
            last_candle.candle.interval = interval;
            last_candle.candle.close_time = trade.ts;

            if is_buy {
                last_candle.candle.buy_taker_amount += trade.amount;
                last_candle.candle.buy_taker_quantity += trade.quantity;
            }

            repo.upsert_candles(&[last_candle]).await?;
        }
    } else {
        insert_new_candle().await?;
    }

    repo.insert_trade(&trade).await?;

    Ok(())
}

async fn _main() -> AppResult<()> {
    dotenv::dotenv().ok();

    let config = get_config()?;
    let poloniex_client = Arc::new(PoloniexClient::new(
        &config.poloniex_rest_url,
        &config.poloniex_ws_url,
    ));
    let repo = Arc::new(SqliteRepo::new_migrate(&config.database_url).await?);
    let mut fetchers = vec![];

    for pair in &config.pairs {
        for interval in &config.intervals {
            let start_time = {
                let last_candle = repo
                    .get_latest_candle_from_interval(&pair, *interval)
                    .await?;

                match last_candle {
                    Some(c) => c.candle.start_time,
                    None => config.start_from,
                }
            };

            let fetcher = fetch_candles_until_now(
                poloniex_client.clone(),
                pair.to_string(),
                *interval,
                start_time,
            );
            fetchers.push(fetcher);
        }
    }
    let fetched_candles = try_join_all(fetchers).await?;

    let candles_to_upsert = fetched_candles
        .into_iter()
        .flat_map(|(candles, pair)| {
            candles
                .into_iter()
                .map(|candle| CandleExtended {
                    candle,
                    pair: pair.clone(),
                })
                .collect::<Vec<_>>()
        })
        .collect::<Vec<CandleExtended>>();

    repo.upsert_candles(&candles_to_upsert).await?;

    let mut trades = poloniex_client.recent_trades_stream(&config.pairs).await?;

    while let Some(t) = trades.next().await {
        let trade = match t {
            Ok(t) => t,
            Err(e) => {
                println!("trade stream error: {e:?}");
                break;
            }
        };
        println!("inserting {trade:?}");

        for interval in &config.intervals {
            tokio::spawn({
                let repo = repo.clone();
                let interval = *interval;
                let trade = trade.clone();
                async move {
                    let result = calculate_new_candles(repo, interval, trade).await;
                    if let Err(e) = result {
                        eprintln!("processor stopped with error: {e:?}")
                    }
                }
            });
        }
    }
    Ok(())
}

#[tokio::main]
async fn main() {
    if let Err(e) = _main().await {
        eprintln!("{e:?}");
    }
}
