use chrono::{DateTime, NaiveDateTime, TimeDelta};
use serde::{Deserialize, Deserializer, Serialize};
use serde_tuple::Deserialize_tuple;
use sqlx::prelude::{FromRow, Type};

#[derive(strum::EnumString, strum::AsRefStr, Clone, Copy, Deserialize, Debug, Type)]
pub enum CandleInterval {
    #[strum(serialize = "MINUTE_1")]
    #[serde(rename = "MINUTE_1")]
    M1,
    #[strum(serialize = "MINUTE_15")]
    #[serde(rename = "MINUTE_15")]
    M15,
    #[strum(serialize = "HOUR_1")]
    #[serde(rename = "HOUR_1")]
    H1,
    #[strum(serialize = "DAY_1")]
    #[serde(rename = "DAY_1")]
    D1,
}

impl CandleInterval {
    pub fn as_timedelta(&self) -> TimeDelta {
        match self {
            CandleInterval::M1 => TimeDelta::minutes(1),
            CandleInterval::M15 => TimeDelta::minutes(15),
            CandleInterval::H1 => TimeDelta::hours(1),
            CandleInterval::D1 => TimeDelta::days(1),
        }
    }
}

#[derive(Debug, Deserialize_tuple, FromRow)]
pub struct Candle {
    #[serde(deserialize_with = "deser_str_to_int")]
    pub low: f64,
    #[serde(deserialize_with = "deser_str_to_int")]
    pub high: f64,
    #[serde(deserialize_with = "deser_str_to_int")]
    pub open: f64,
    #[serde(deserialize_with = "deser_str_to_int")]
    pub close: f64,
    #[serde(deserialize_with = "deser_str_to_int")]
    pub amount: f64,
    #[serde(deserialize_with = "deser_str_to_int")]
    pub quantity: f64,
    #[serde(deserialize_with = "deser_str_to_int")]
    pub buy_taker_amount: f64,
    #[serde(deserialize_with = "deser_str_to_int")]
    pub buy_taker_quantity: f64,
    pub trade_count: i64,
    #[serde(deserialize_with = "deser_naive_dt")]
    pub ts: NaiveDateTime,
    #[serde(deserialize_with = "deser_str_to_int")]
    pub weighted_average: f64,
    pub interval: CandleInterval,
    #[serde(deserialize_with = "deser_naive_dt")]
    pub start_time: NaiveDateTime,
    #[serde(deserialize_with = "deser_naive_dt")]
    pub close_time: NaiveDateTime,
}

#[derive(FromRow)]
pub struct CandleExtended {
    #[sqlx(flatten)]
    pub candle: Candle,
    pub pair: String,
}

fn deser_str_to_int<'de, D: Deserializer<'de>>(deserialize: D) -> Result<f64, D::Error> {
    let s = String::deserialize(deserialize)?;
    s.parse()
        .map_err(|e| serde::de::Error::custom(format!("{e}: {s}")))
}

fn deser_naive_dt<'de, D: Deserializer<'de>>(deserialize: D) -> Result<NaiveDateTime, D::Error> {
    let ts = Deserialize::deserialize(deserialize)?;
    DateTime::from_timestamp_millis(ts)
        .ok_or(serde::de::Error::custom(format!("wrong timestamp '{ts}'")))
        .map(|dt| dt.naive_utc())
}

#[derive(Deserialize, Debug, Clone)]
#[serde(rename_all = "camelCase")]
pub struct Trade {
    pub symbol: String,
    #[serde(deserialize_with = "deser_str_to_int")]
    pub amount: f64,
    pub taker_side: TradeDirection,
    #[serde(deserialize_with = "deser_str_to_int")]
    pub quantity: f64,
    #[serde(deserialize_with = "deser_naive_dt")]
    pub create_time: NaiveDateTime,
    #[serde(deserialize_with = "deser_str_to_int")]
    pub price: f64,
    pub id: String,
    #[serde(deserialize_with = "deser_naive_dt")]
    pub ts: NaiveDateTime,
}

#[derive(Deserialize, Debug, strum::AsRefStr, Clone, Copy, Type)]
#[serde(rename_all = "camelCase")]
#[sqlx(rename_all = "camelCase")]
pub enum TradeDirection {
    Buy,
    Sell,
}

#[derive(Serialize)]
pub struct SubscriptionRequest {
    pub event: String,
    pub channel: Vec<String>,
    pub symbols: Vec<String>,
}

#[derive(Deserialize, Debug)]
#[allow(dead_code)]
pub struct SubscriptionResponse {
    pub event: String,
    pub channel: String,
    pub symbols: Vec<String>,
}

#[derive(Deserialize, Debug)]
#[allow(dead_code)]
pub struct SubscriptionResponseData<T> {
    pub channel: String,
    pub data: Vec<T>,
}
