use std::str::FromStr;

use sqlx::{sqlite::SqliteConnectOptions, SqlitePool};

use crate::{
    error::{AppError, AppResult},
    models::{CandleExtended, CandleInterval, Trade},
};

use super::Repo;

pub struct SqliteRepo {
    pool: SqlitePool,
}

impl SqliteRepo {
    pub async fn new_migrate(db_url: &str) -> AppResult<Self> {
        let conn_options = SqliteConnectOptions::from_str(db_url)?.create_if_missing(true);

        println!("trying to open db {db_url}");
        let pool = SqlitePool::connect_with(conn_options).await?;

        println!("applying migrations");
        sqlx::migrate!("./migrations").run(&pool).await?;

        Ok(Self { pool })
    }
}

#[async_trait]
impl Repo for SqliteRepo {
    async fn upsert_candles(&self, candles: &[CandleExtended]) -> AppResult<u64> {
        let q = "
            REPLACE INTO candles(
                low,
                high,
                open,
                close,
                amount,
                quantity,
                buy_taker_amount,
                buy_taker_quantity,
                trade_count,
                ts,
                weighted_average,
                interval,
                start_time,
                close_time,
                pair
            ) VALUES (
                $1,
                $2,
                $3,
                $4,
                $5,
                $6,
                $7,
                $8,
                $9,
                $10,
                $11,
                $12,
                $13,
                $14,
                $15
            )
        ";

        let mut tx = self.pool.begin().await?;
        let mut affected = 0;
        for candle in candles {
            affected += sqlx::query(q)
                .bind(&candle.candle.low)
                .bind(&candle.candle.high)
                .bind(&candle.candle.open)
                .bind(&candle.candle.close)
                .bind(&candle.candle.amount)
                .bind(&candle.candle.quantity)
                .bind(&candle.candle.buy_taker_amount)
                .bind(&candle.candle.buy_taker_quantity)
                .bind(&candle.candle.trade_count)
                .bind(&candle.candle.ts)
                .bind(&candle.candle.weighted_average)
                .bind(&candle.candle.interval)
                .bind(&candle.candle.start_time)
                .bind(&candle.candle.close_time)
                .bind(&candle.pair)
                .execute(&mut *tx)
                .await?
                .rows_affected();
        }

        tx.commit().await?;
        Ok(affected)
    }

    async fn insert_trade(&self, trade: &Trade) -> AppResult<u64> {
        let q = "
            INSERT INTO trades(
                symbol,
                amount,
                taker_side,
                quantity,
                create_time,
                price,
                id,
                ts
            ) VALUES (
                $1,
                $2,
                $3,
                $4,
                $5,
                $6,
                $7,
                $8
            );
        ";

        sqlx::query(&q)
            .bind(&trade.symbol)
            .bind(&trade.amount)
            .bind(&trade.taker_side)
            .bind(&trade.quantity)
            .bind(&trade.create_time)
            .bind(&trade.price)
            .bind(&trade.id)
            .bind(&trade.ts)
            .execute(&self.pool)
            .await
            .map_err(AppError::from)
            .map(|r| r.rows_affected())
    }

    async fn get_latest_candle_from_interval(
        &self,
        pair: &str,
        interval: CandleInterval,
    ) -> AppResult<Option<CandleExtended>> {
        let q = "
            SELECT * FROM candles
            WHERE pair = $1 AND interval = $2
            ORDER BY start_time DESC
            LIMIT 1
        ";

        sqlx::query_as::<_, CandleExtended>(q)
            .bind(pair)
            .bind(interval)
            .fetch_optional(&self.pool)
            .await
            .map_err(AppError::from)
    }
}
